# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

import logging
import os
import stat
import time

from cmf.monitor import schema
from cmf.monitor.abstract_monitor import CollectionError
from cmf.monitor.abstract_hdfs_monitor import AbstractHdfsMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.role_update import RoleUpdate
from cmf.monitor.service_update import ServiceUpdate
import cmf.monitor.namenode.json_parser
import cmf.monitor.namenode.json_parser_jmx
from cmf.monitor.json_metrics_utils import add_derived_jvm_metrics_metrics, \
    add_derived_jvm_metrics_jmx, get_metrics_metric, add_metrics_metrics, \
    add_metrics_jmx
from cmf.monitor.jmx_beans import NAMENODE_CLIENT_RPC_PORT_NAME, \
    NAMENODE_SERVICE_RPC_PORT_NAME
from cmf.monitor.cgroup_metrics_utils import add_cgroup_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics
from cmf.monitor import constants
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger('NameNodeMonitor')
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

_ROLETYPE = "NAMENODE"
_HDFS_TYPE = "HDFS"

# Section and key names in cloudera-monitor.properties
# Since CM 3.7
_SECTION = 'namenode'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_DFS_HTTP_PORT = 'dfs_http_port'
_NAMENODE_CLIENT_RPC_PORT = 'namenode_port'
_DFS_NAME_DIRS = 'dfs_name_dirs'
_LOG_DIR = 'log_dir'
_COLLECT_INTERVAL = 'collect_interval'
# Added in CM 4.0
_SERVICE_VERSION = 'service_version'
# Added in CM 4.6.2
# This is the service RPC port -- the port to which HDFS service members talk.
# This may not be set.
_NAMENODE_SERVICE_RPC_PORT = 'dfs_namenode_servicerpc_address'

def get_fsnamesystem_metric(metrics, metric, default=None):
  return get_metrics_metric(metrics, 'dfs', 'FSNamesystem', metric, default)

def get_namenode_metric(metrics, metric, default=None):
  return get_metrics_metric(metrics, 'dfs', 'namenode', metric, default)

class NameNodeMonitor(AbstractHdfsMonitor):

  def __init__(self, daemon, registry, cgroup_root=None, cgroup_name=None):
    AbstractHdfsMonitor.__init__(self,
                                 daemon,
                                 registry,
                                 cgroup_root,
                                 cgroup_name,
                                 _SECTION,
                                 _DFS_HTTP_PORT)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_role_report_timestamp = None
    self._last_service_report_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]
    self._metric_accessors = daemon.metric_accessors

    # Config data
    self._service_name = None
    self._service_version = None
    self._name = None
    self._dfs_name_dirs = [ ]
    self._log_dir = None
    self._metric_lookup_context = { }

  def __str__(self):
    return "NameNodeMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Overrides AbstractMonitor. Returns True/False for success"""
    try:
      service_name = confparse.get(_SECTION, _SERVICE_NAME)
      service_version = \
          confparse.getint_with_default(_SECTION, _SERVICE_VERSION,
                                        constants.SERVICE_VERSION_CDH3)
      name = confparse.get(_SECTION, _ROLE_NAME)
      dfs_name_dirs = confparse.getlist(_SECTION, _DFS_NAME_DIRS)
      client_rpc_port = confparse.getint(_SECTION, _NAMENODE_CLIENT_RPC_PORT)
      log_dir = confparse.get(_SECTION, _LOG_DIR)
      host = confparse.get(_SECTION, _HOST)
      web_port = confparse.get(_SECTION, _DFS_HTTP_PORT)
      service_rpc_port = \
          confparse.getint_with_default(_SECTION,
                                        _NAMENODE_SERVICE_RPC_PORT,
                                        None)
    except ValueError, ex:
      self.record_error(ex)
      return False

    metrics_url = self._compute_metrics_url(service_version, host, web_port)
    if metrics_url is None:
      return False

    self._service_name = service_name
    self._service_version = service_version
    self._name = name
    self._dfs_name_dirs = dfs_name_dirs
    self._log_dir = log_dir
    self._metrics_url = metrics_url

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    self.create_dir_monitor(self._log_dir, self._role_dir_monitors)
    for d in self._dfs_name_dirs:
      self.create_dir_monitor(d, self._role_dir_monitors)

    # Build the metric lookup context. This always needs to contain the client
    # RPC port and may contain the service RPC port if it is set.
    self._metric_lookup_context = dict()
    self._metric_lookup_context[NAMENODE_CLIENT_RPC_PORT_NAME] = client_rpc_port
    if service_rpc_port is not None:
      self._metric_lookup_context[NAMENODE_SERVICE_RPC_PORT_NAME] = \
          service_rpc_port

    return True

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_role_report_timestamp,
                                       update):
      self._last_role_report_timestamp = update.timestamp

    add_proc_metrics(update, self._pid, LOG)
    add_cgroup_metrics(update, self._cgroup_root, self._cgroup_name)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    # For CDH3 we get the time since the last checkpoint by looking at files
    # on the filesystem. For CDH4 we get this information from JMX metrics.
    if self._service_version == constants.SERVICE_VERSION_CDH3 or \
       self._service_version == constants.SERVICE_VERSION_CDH3U4X:
      self._add_time_since_last_checkpoint(update)

    self.add_unexpected_exit_metrics(update)

    return update

  def get_service_update(self):
    # The HDFS service update is only returned for CDH 3. CDH 4 and beyond have
    # the metrics generated off the HDFS poller inside SMON instead.
    if self._service_version != constants.SERVICE_VERSION_CDH3 and \
       self._service_version != constants.SERVICE_VERSION_CDH3U4X:
      return None

    (metrics, timestamp, update) = self._metrics_tuple
    if metrics is None or isinstance(metrics, CollectionError):
      return None

    # We don't want to double report the same set of metrics.
    if timestamp == self._last_service_report_timestamp:
      return None
    self._last_service_report_timestamp = timestamp

    update = ServiceUpdate(name=self._service_name,
                           service_type=_HDFS_TYPE,
                           timestamp=timestamp)

    update.add_metric(schema.BLOCKS_TOTAL,
                      get_fsnamesystem_metric(metrics, 'BlocksTotal'))
    # See OPSAPS-6070: 'corruptBlocks' in CDH3 means corrupt replicas.
    update.add_metric(schema.CORRUPT_REPLICAS,
                      get_fsnamesystem_metric(metrics, 'CorruptBlocks'))
    update.add_metric(schema.MISSING_BLOCKS,
                      get_fsnamesystem_metric(metrics, 'MissingBlocks'))
    update.add_metric(schema.UNDER_REPLICATED_BLOCKS,
                      get_fsnamesystem_metric(metrics, 'UnderReplicatedBlocks'))
    update.add_metric(schema.PENDING_DELETION_BLOCKS,
                      get_fsnamesystem_metric(metrics, 'PendingDeletionBlocks'))
    update.add_metric(schema.TOTAL_LOAD,
                      get_fsnamesystem_metric(metrics, 'TotalLoad'))
    update.add_metric(schema.PENDING_REPLICATION_BLOCKS,
                      get_fsnamesystem_metric(metrics, 'PendingReplicationBlocks'))
    update.add_metric(schema.SCHEDULED_REPLICATION_BLOCKS,
                      get_fsnamesystem_metric(metrics, 'ScheduledReplicationBlocks'))
    update.add_metric(schema.EXCESS_BLOCKS,
                      get_fsnamesystem_metric(metrics, 'ExcessBlocks'))
    update.add_metric(schema.BLOCK_CAPACITY,
                      get_fsnamesystem_metric(metrics, 'BlockCapacity'))
    update.add_metric(schema.FILES_TOTAL,
                      get_fsnamesystem_metric(metrics, 'FilesTotal'))
    
    return update

  def add_webserver_metrics_to_role_update(self, update, metrics):
    if self._service_version == constants.SERVICE_VERSION_CDH3 or \
       self._service_version == constants.SERVICE_VERSION_CDH3U4X:
      self._add_webserver_metrics_to_role_update_cdh3(update, metrics)
    elif self._service_version == constants.SERVICE_VERSION_CDH4:
      self._add_webserver_metrics_to_role_update_cdh4(update, metrics)
    else:
      THROTTLED_LOG.warn("Unknown namenode version: %s" % (self._service_version))

  def _add_webserver_metrics_to_role_update_cdh3(self, update, metrics):
    add_derived_jvm_metrics_metrics(update, metrics)
    add_metrics_metrics(update, metrics, self._metric_accessors, _ROLETYPE, 'cdh3')

  def _add_webserver_metrics_to_role_update_cdh4(self, update, metrics):
    add_derived_jvm_metrics_jmx(update, metrics)
    add_metrics_jmx(update, metrics, self._metric_accessors, _ROLETYPE, 'cdh4')

  def _add_time_since_last_checkpoint(self, update):
    # We use a local reference in case the list of directories is changed out 
    # from under us.
    age = -1
    now = time.time()
    dirs = self._dfs_name_dirs

    for d in dirs:
      p = os.path.join(d, 'current', 'fsimage')
      try:
        age = max(age, now - os.stat(p)[stat.ST_MTIME])
      except (Exception), e:
        LOG.error("Failed to stat %s: %s" % (p, e))

    if age != -1:
      update.add_metric(schema.NN_TIME_SINCE_LAST_CHECKPOINT, age)

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  def collect(self):
    """Overrides AbstractMonitor. Collect NN metrics. Invoked asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      self._metrics_tuple = (None, None, None)
      return

    LOG.debug("Fetching NameNode metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def parse_metrics(self, json):
    if self._service_version == constants.SERVICE_VERSION_CDH3 or \
       self._service_version == constants.SERVICE_VERSION_CDH3U4X:
      return cmf.monitor.namenode.json_parser.JsonParser(json)
    elif self._service_version == constants.SERVICE_VERSION_CDH4:
      return cmf.monitor.namenode.json_parser_jmx.JsonParser(
        json, self._metric_lookup_context)
    raise Exception("Unsupported operation.")

  def get_subject_type(self):
    return _ROLETYPE

  # For testing. 
  def clear_last_reported_timestamp(self):
    super(NameNodeMonitor, self).clear_last_reported_timestamp()
    self._last_role_report_timestamp = None
    self._last_service_report_timestamp = None
